#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <syslog.h>
#include <unistd.h>

using namespace std;

int debug_nofork; 

string read_message(int fd)
{
	int rslt, sz = 4096;
	string s = "";
	char buf[sz];
	
	long msg_len;
	rslt = read(fd, (void*) &msg_len, sizeof(long));
	msg_len = ntohl(msg_len);
	
	if(debug_nofork) {
		syslog(LOG_INFO, "Read: expect to read %ld", msg_len);
	}	
	
	while(1) {
	
		if(debug_nofork) {
			syslog(LOG_INFO, "Read: in while 1, s is %s, msg_len is %ld", s.c_str(), msg_len);
		}
	
		memset(buf, 0, sz);
		rslt = read(fd, buf, msg_len);
		if(rslt == 0) {
			break;
		}
		if(rslt > 0 ) {
			s += buf;
			msg_len -= rslt;
		} else {
			if(errno != EINTR) {
				perror("Error on fd read: ");
				break;
			}
		}
		
		if(msg_len <= 0) {
			break;
		}
		
	} // while
	
	if(debug_nofork) {
		syslog(LOG_INFO, "Read: returning");
	}
	
	return s;
}

int pipe_sendmail_message(const string& msg)
{
	int write_rslt, rslt = 0;
	FILE* fp = popen("sendmail -t", "w");
	write_rslt = fwrite(msg.c_str(), sizeof(char), msg.length(), fp);
	pclose(fp);
	return rslt;	
}

int accept_handle_client(int& listenfd) 
{
	int child_pid, connfd, rslt, status;
	struct sockaddr_un cliaddr;
	
	rslt = 0;
	socklen_t clilen = sizeof(cliaddr);
	connfd = accept(listenfd, (struct sockaddr*) &cliaddr, &clilen);
	if(connfd < 0) {
		if(errno != EINTR) {
			/* probably should throw here */
		}
		rslt = errno;
		goto accept_handle_client_out;
	}
	
	/* Got it */
	
	if(debug_nofork) {
		/* ja rule */
		string msg = read_message(connfd);
		pipe_sendmail_message(msg);
	} else {
		/* fork me */
		child_pid = fork();
		if(child_pid == 0) {
			/* Child */
			close(listenfd);
			string msg = read_message(connfd);
			pipe_sendmail_message(msg);
			exit(0);
		} else {
			close(connfd);
			waitpid(child_pid, &status, 0);
		}
	} // fork

accept_handle_client_out:
	return rslt;
}

int real_main(int argc, char** argv)
{
	int rslt = 0;
	int listenfd;
	struct sockaddr_un srvaddr;
		
	const string socket_path = "/var/run/fv_milter_queued.sock";
	const int listen_backlog = 200;
	
	try {
		listenfd = socket(PF_UNIX, SOCK_STREAM, 0);
	
		/* Must remove old file, if any */
		unlink(socket_path.c_str());
		memset(&srvaddr, 0, sizeof(srvaddr));
	
		/* Is this safe? */
		srvaddr.sun_family = PF_UNIX;
		strcpy(srvaddr.sun_path, socket_path.c_str());
		socklen_t srvlen = sizeof(srvaddr);
	
		/* Setup */
		int bind_rslt = bind(listenfd, (struct sockaddr*) &srvaddr, srvlen);
		if(bind_rslt != 0) {
			perror("Error creating listening socket: ");
			rslt = 1;
			throw(0);
		}
		
		listen(listenfd, listen_backlog);
	
		/* Big Loop */
		while(1) {	
			int cli_rslt = accept_handle_client(listenfd);
			if(cli_rslt != 0) {
				/* do something about it? */
			}
		}
	} // try
	
	catch(...) {
		// do something?
	}
	
	return rslt;
}

int main(int argc, char** argv)
{
	// We need to become a daemon process -- this
	// is UNIX boilerplate

	if( argc >= 2 ) // unless otherwise specified
	{
		if( strcmp(argv[1], "--nofork") == 0 )
		{
			debug_nofork = 1;
			syslog(LOG_INFO, "Started with --nofork, not forking");
			return real_main( argc, argv );
		}
	}

	int pid = fork();
	if(pid == 0)
	{
		setsid();
		pid = fork();

		if(pid == 0)
		{
			return real_main( argc, argv);
		}
	}
	
} // main

